package org.ucam.srcf.assassins;

import java.util.Collection;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Comparator;
import java.util.Random;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Iterator;
import java.io.PrintWriter;
import java.io.FileOutputStream;
import java.io.File;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.ObjectModifiedException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.xml.ColourFilter;
import org.ucam.srcf.assassins.xml.Config;

public abstract class PageCreator implements Comparator{
   protected String target; // The page to write to
   private TreeSet allNewsPages;
   protected SortedSet events;
   protected PrintWriter out;
   private Map eventState;
   protected Map colourMaps = new HashMap();
   protected Map references;
   protected Map psuedonyms;
   private String linkName="";
   private Map plugins;
   
   public PageCreator(String t, Map p, Comparator c) throws java.io.IOException {
      plugins = p;
      events = new TreeSet(c);
      eventState = new HashMap();
      references = new HashMap();
      target = t;
      out = new PrintWriter(new FileOutputStream(new File(target)));
      allNewsPages= new TreeSet(this);
   }

   public int compare(Object arg0, Object arg1) {
      if (Debug.debug) Debug.print(PageCreator.class, "Arg0: "+arg0);
      if (Debug.debug) Debug.print(PageCreator.class, " Arg1: "+arg1);
      if (Debug.debug) Debug.print(PageCreator.class, " Arg0.getLinkName(): "+((PageCreator)arg0).getLinkName());
      if (Debug.debug) Debug.print(PageCreator.class, " Arg1.getLinkName(): "+((PageCreator)arg1).getLinkName());
      PageCreator a = (PageCreator) arg0;
      PageCreator b = (PageCreator) arg1;
      return a.linkName.compareTo(b.linkName);
      //return ((PageCreator)arg0).getLinkName().compareTo(((PageCreator)arg1).getLinkName());
   }


   public void addEvent(Event x, Map states, PageCreator ref) throws ObjectModifiedException, AssassinSQLException {
      events.add(x);
      eventState.put(x,states);
      references.put(x,ref);
      if (Debug.debug) Debug.print(PageCreator.class, "Added event, eid="+x.getEid());
   }

   public String getLinkName()
   {
      return linkName;
   }

   public void setLinkName(String ln)
   {
      linkName = ln;
   }
   
   public Map updatePseudonymMap(Map m, Integer n, Assassin a, Pseudonym p, Event e) throws ObjectModifiedException, AssassinSQLException, StateBuilderException {
      String pxt;
      String myRealName;
      Player pt = null;
      if (null==a) {
         return m;
         // myRealName = "ERROR! NO ASSASSIN GIVEN!";
      }
      pt = a.getPlayer();
      if (null==pt) {
         return m;
         // myRealName="ERROR! NO PLAYER ASSOCIATED!";
      }
      myRealName = pt.getName();
      if (null==p) {
         pxt = myRealName;
      } else {
         pxt = p.getValue();
      }
      Map currentState = (Map) eventState.get(e);
      AssassinState as = (AssassinState) currentState.get(a);
      List las;
      if (null==as) {
         // this is definitely an error condition. execution should not
         // continue past this point.  (amgb2)
         //
         // actually, I think this occurs when you have a report by someone who
         // is not a participant in the event. It should probably do something 
         // sensible (mjj29)
         //
         // I have made it generate the state for this assassin at the event time
         LightweightStateBuilder lsb = new LightweightStateBuilder(e.getEventTime());
         lsb.registerInterest(a);
         as = (AssassinState) lsb.generateState().get(a);
      }
      las = as.getPseudonyms();
      String pscolour = getColourForAssassin(as, pxt);
      String lassoo = ""; // Mike, 'lassoo'? wtf? -- amgb2
      String colourlassoo = "";
      Iterator lassit = las.iterator();
      if (Debug.debug) Debug.print(PageCreator.class, "begin iteration");
      while (lassit.hasNext()) {
         String nextP = Tools.escapeXML(((Pseudonym)lassit.next()).getValue());
         lassoo += (lassoo.equals("")?"":" AKA ") + nextP;
         colourlassoo += (colourlassoo.equals("")?"":" AKA ") + "<colour type=\""+pscolour+"\">" + nextP + "</colour>";
      }
      if (Debug.debug) Debug.print(PageCreator.class, "end iteration");
      //String rncolour = getColourForAssassin(as, myRealName);
      //String pscolour = getColourForAssassin(as, pxt);
      
      Map eventColourMap = (Map) colourMaps.get(e);
      if (null == eventColourMap) eventColourMap = new HashMap();
      eventColourMap.put(a, pscolour);
      colourMaps.put(e, eventColourMap);
      
      if (Debug.debug) Debug.print(PageCreator.class, "lassoo == " + lassoo);
      //String colourlassoo = "<colour type=\""+pscolour+"\">"+Tools.replace(lassoo, (Map) (new HashMap()).put(" AKA ","</colour> AKA <colour type=\""+pscolour+"\">"))+"</colour>";
      if (Debug.debug) Debug.print(PageCreator.class, "colourlassoo == " + colourlassoo);
      m.put("P"+n,"<colour type=\""+pscolour+"\">"+Tools.escapeXML(pxt)+"</colour>");
      m.put("p"+n,Tools.escapeXML(pxt));
      m.put("N"+n,"<colour type=\""+pscolour+"\">"+Tools.escapeXML(myRealName)+"</colour>");
      m.put("n"+n,Tools.escapeXML(myRealName));
      m.put("L"+n,colourlassoo);
      m.put("l"+n,lassoo);
      return m;
   }

   private String getColourForAssassin(AssassinState as, String name)
   {
      String colourClass = getColourClassForAssassin(as);
      if (null == colourClass) return "unknown";
      int range = Config.getConfigInt("vanilla", "colourcount-" + colourClass);
      if (range == 0) return "unknown";
      int suffix = 1 + (new Random(name.hashCode())).nextInt(range);
      return colourClass + suffix;
   }
   
   private String getColourClassForAssassin(AssassinState as)
   {
      Map classes = Config.getColourFilters();
      Iterator i = classes.keySet().iterator();
      if (Debug.debug) Debug.print(PageCreator.class, "finding match");
      while (i.hasNext())
      {
         String classKey = (String) i.next();
         ColourFilter cf = (ColourFilter) classes.get(classKey);
         if (Debug.debug) Debug.print(PageCreator.class, " trying " + classKey);
         if (cf.matchAssassinState(plugins, as)) return classKey;
      }
      if (Debug.debug) Debug.print(PageCreator.class, " failure");
      return null;
   }
  
   public abstract void generate () throws AssassinSQLException, StateBuilderException;

   /**
    * @param collection
    */
   public void setAllNewsPages(Collection pages) {
   	  this.allNewsPages.clear();
   	  if (pages!=null) {
   	  	for (Iterator i = pages.iterator(); i.hasNext(); ) {
   	  		PageCreator pc = (PageCreator) i.next();
   	  		if (null!=pc) this.allNewsPages.add(pc);
   	  	}
   	  }
   }
   
   public Set getAllNewsPages() {
   	return this.allNewsPages;
   }
}
