/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.chocolate;
import java.text.ParseException;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
import java.util.HashMap;

import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.ConfigProperty;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.PlayerViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginCommand;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.TextQuestion;
import org.ucam.srcf.assassins.Tools;
import org.ucam.srcf.assassins.YesNoQuestion;
import org.ucam.srcf.assassins.autoumpire;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Chocolate extends Plugin
{
   public static String getDescription() { return "Tracks the competency of players"; };
   private static final long DAYSMILI = 1000*60*60*24;
   private int defaultDays;
   private int warningDays;
   private long startTime;
   private String deadlineString;
   private String incoString;

   public void setPluginName(String s) 
   {
      super.setPluginName(s);
      deadlineString = Config.getConfig(getPluginName(), "deadline-at");
      incoString = Config.getConfig(getPluginName(), "incompetent");
      defaultDays = Integer.parseInt(Config.getConfig(getPluginName(), "default-extension"));
      warningDays = Integer.parseInt(Config.getConfig(getPluginName(), "warning"));
      try {
         startTime = Tools.unFormatTime(Config.getConfig("vanilla", "game-start-date"));
      } catch (ParseException pe) { startTime = 0; }
   }
   
   /**
    * Called whenever the event to which an existing state change is
    * attached has its time changed.
    * @param sc The state change affected.
    * @param oldtime The previous timing of the state change's associated
    *    event.
    * @param newtime The new timing of the event. Note that this can also
    *    be retrieved with {@link StateChange#getEvent} and
    *    {@link Event#getTime}, but is provided here for convenience.
    */
   protected void stateChangeTimeChanged(StateChange sc, int oldtime, int newtime)  throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(autoumpire.class, "stateChangeTimeChanged("+sc+","+oldtime+","+newtime+")");
      if (!sc.getPluginName().equals(getPluginName())) return;
      if (!sc.getFieldName().equals("competency")) return;

      AssassinViewer av = sc.getAssassinViewer();
      long time = Long.parseLong(sc.getStringValue());
      time += (newtime - oldtime);
      EventViewer ev = sc.getEventViewer();
      vanilla.removeStateChange(ev, sc);
      vanilla.addStateChange(ev, new PluginStateChange(av, this, "competency", "" + time));
   }
   
   /**
    * Called whenever the plugin is required to update or sanity-check the
    * current gamestate.
    */
   protected Vector updateState(Map state, Vector qs, EventViewer e) throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(Chocolate.class, "updateState("+state+","+qs+")");
      Vector v = new Vector();
      Iterator i = state.keySet().iterator();
      int j = 0;
      while(i.hasNext()) {
         AssassinViewer a = (AssassinViewer) i.next();
         if (Debug.debug) Debug.print(Chocolate.class, "state.containsKey(a) = " + state.containsKey(a));
         if (Debug.debug) Debug.print(Chocolate.class, "assassin = " + a);      
         if (Debug.debug) Debug.print(Chocolate.class, "a.equals(null) = " + a.equals(null));
         PlayerViewer p = a.getPlayerViewer();
         if (Debug.debug) Debug.print(Chocolate.class, "operating finches");
         AssassinStateViewer asv = (AssassinStateViewer) state.get(a);
         if (Debug.debug)
         {
            if (null == asv) Debug.print(Chocolate.class, "warning program about to crash");
         }
         if (Debug.debug) Debug.print(Chocolate.class, "got viewers");
         boolean competent = asv.getBoolProperty(this, "competent");
         if (Debug.debug) Debug.print(Chocolate.class, "determined competence");
         String deadstring = asv.getProperty(this, "competency");
         long deadline;
         if (null == deadstring) {
            deadline = startTime + defaultDays * DAYSMILI;
         }
         else {
            deadline = Long.parseLong(deadstring);
         }
         if (Debug.debug) Debug.print(Chocolate.class, "extracted properties");
         long now = System.currentTimeMillis();
         if (competent && deadline < now && asv.isActive() && asv.isAlive()) {
            if (0 != qs.size()) {
              YesNoQuestion q = (YesNoQuestion) qs.get(j++);
              if (q.getAnswer()) {
                 vanilla.addStateChange(e, new PluginStateChange(a, this, "competent", "false"));
              }
            }
            else {
               if (Debug.debug) Debug.print(Chocolate.class, "adding question");
               v.add(new YesNoQuestion("Make "+p.getName()+" Incompetent?", true));
            }
         }
      }
      if (0 == v.size()) return null;
      return v;
   }
   
   /**
    * Called by the filtering mechanism when the plugin needs to determine
    * whether an assassin has a given property such as &quot;police&quot;. If
    * the plugin does not recognise the named property, it should return
    * <code>false</code>.
    * @param a A viewer for the assassin being tested.
    * @param prop The name of the property being asked about.
    * @return True if the assassin has the property, false if not.
    */
   public boolean assassinHasProperty(AssassinStateViewer asv, String prop, String value)
   {
      if (Debug.debug) Debug.print(autoumpire.class, "assassinHasProperty("+asv+","+prop+","+value+")");
      if (!"competent".equals(prop)) return false;
      String competent = asv.getProperty(this, "competent");
      if (null == competent) return false;
      if (competent.equals(value)) return true;
      return false;
   }

   /**
    * Called during the updating of the gamestate to ask whether a given player
    * requires email to be sent to them during the current round of emails.
    * @param p A viewer for the player being asked about.
    * @return True if the plugin requires email to be sent to the player in
    *    question, false if not.
    */
   protected boolean playerNeedsEmail(AssassinStateViewer asv) 
   {
      if (Debug.debug) Debug.print(autoumpire.class, "playerNeedsEmail("+asv+")");
      boolean competent = asv.getBoolProperty(this, "competent");
      String s = asv.getProperty(this, "competency");
      if (null == s) return false;
      long deadline = Long.parseLong(s);
      long now = System.currentTimeMillis();
      if (competent && deadline + (warningDays*DAYSMILI) < now)
         return true;
      return false;
   }

   /**
    * Called when the system has decided that email needs to be sent to a given
    * player. The plugin may return a string to be included in the email, or
    * null if it does not wish to include anything.
    * @param p A viewer for the player being emailed.
    * @return A string to be included in the email sent to the player, or null.
    *    The string will appear in its own paragraph when inserted into the
    *    email and so should not begin or end with a newline.
    */
   protected String generateEmail(AssassinStateViewer asv, Map state) throws ParseException 
   {
      if (Debug.debug) Debug.print(autoumpire.class, "generateEmail("+asv+")");
      boolean competent = asv.getBoolProperty(this, "competent");
      String s = asv.getProperty(this, "competency");
      long deadline;
      if (null == s) deadline = startTime + defaultDays * DAYSMILI;
      else deadline = Long.parseLong(s);
      if (!competent)
         return incoString;
      return deadlineString + Tools.formatTime(deadline);
   }

   /**
    * Called when an event is being added to get any queries the plugin
    * has for the user. The plugin should check all the event details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param e The Event which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector getQuestions(EventViewer e, Map state) throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(autoumpire.class, "getQuestions("+e+","+state+")");
      Vector v = new Vector();
      Iterator i = e.getParticipants().values().iterator();
      while(i.hasNext()) {
         AssassinViewer a = (AssassinViewer) i.next();
         PlayerViewer p = a.getPlayerViewer();
         AssassinStateViewer as = (AssassinStateViewer) state.get(a);
         if (as.getBoolProperty(this, "competent"))
            v.add(new YesNoQuestion("Update competency for "+p.getName()+"?", true));
         else
            v.add(new YesNoQuestion("Redeem " + p.getName() + " from incompetency?", true));
         v.add(new TextQuestion("Competency Extension (days):", "" + defaultDays));
      }
      return v;

    }
   
   /**
    * Returns the Questions with the answers filled out
    *
    * @param e The Event being added
    * @param qs The Questions which were asked
    */
   protected void returnQuestions(EventViewer e, Map state, Vector qs)  throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(autoumpire.class, "returnQuestions("+e+","+state+","+qs+")");
      Map participants = e.getParticipants();
      Iterator i = participants.values().iterator();
      int j = 0;
      while (i.hasNext()) {         
         boolean competency = ((YesNoQuestion) qs.get(j)).getAnswer();
         int extension = Integer.parseInt(((TextQuestion) qs.get(j+1)).getAnswer());
         AssassinViewer a = (AssassinViewer) i.next();
         AssassinStateViewer as = (AssassinStateViewer) state.get(a);         
         if (competency) {
            vanilla.addStateChange(e, new PluginStateChange(a, this, "competent", "true"));
            String s = as.getProperty(this, "competency");
            long oldCompetency;
            if (null == s) oldCompetency = 0;
            else oldCompetency = Long.parseLong(s);
            long newCompetency = e.getTime() + (extension*DAYSMILI);
            if (Debug.debug) Debug.print(autoumpire.class, "Updating competency for " + a.getPlayerViewer().getName() + " to " + newCompetency);
            if (newCompetency > oldCompetency)
               vanilla.addStateChange(e, new PluginStateChange(a, this, "competency", ""+newCompetency));
         }
         j+=2;
      }
   }

   /**
    * Called when an assassin is being added to get any queries the plugin
    * has for the user. The plugin should check all the assassin details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param a The Assassin which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector getQuestions(AssassinViewer a, Map state) throws AssassinSQLException 
   {
      Vector qs = new Vector();
      qs.add(new TextQuestion("Initial Competence Period (Days):","7"));
      return qs;
   }
   
   /**
    * Returns the Questions with the answers filled out
    *
    * @param a The Assassin being added
    * @param qs The Questions which were asked
    */ 
   protected void returnQuestions(AssassinViewer a, EventViewer e, Map state, Vector qs) throws AssassinSQLException
   {
      String period = ((TextQuestion)qs.get(0)).getAnswer();
      long basetime = e.getTime();
      if (basetime < startTime) basetime = startTime;
      long deadline = basetime + (Integer.parseInt(period) * DAYSMILI);
      vanilla.addStateChange(e, new PluginStateChange(a, this, "competent", "true"));
      vanilla.addStateChange(e, new PluginStateChange(a, this, "competency", "" + deadline));
   }
   public void dumpDeadlines(Map params) throws AssassinSQLException, StateBuilderException {
      Map state = vanilla.getState();
	   for (Iterator i = state.keySet().iterator(); i.hasNext();) {
		   AssassinViewer a = (AssassinViewer) i.next();
		   AssassinStateViewer as = (AssassinStateViewer) state.get(a);
		   if (as.isAlive() && as.isActive() && (as.getBoolProperty(getPluginName(), "competent") != params.containsKey("-i"))) {
			   System.out.println(a.getPlayerViewer().getName() + "," + Tools.formatTime(Long.parseLong(as.getProperty(getPluginName(), "competency"))));
		   }
	   }
   }
   public Map getCommands() {
	   Map m = new HashMap();
	   final Chocolate c = this;
	   m.put("dump-deadlines",
			   new PluginCommand() {
				   public void pluginCommand(Map params) throws AssassinSQLException, StateBuilderException {
					   c.dumpDeadlines(params);
				   }
			   }
		);
	   return m;
   }
   
   public static Vector getConfigProperties() 
   { 
      Vector v = new Vector(); 
      v.add ( new ConfigProperty ("deadline-at", "String to print competence deadline in emails", "Your competence deadline is at: "));
      v.add ( new ConfigProperty ("incompetent", "String to print in email if you are incompetent", "You are currently marked as incompetent. I suggest you get out there and do something about it!"));
      v.add ( new ConfigProperty ("default-extension", "Default competence extension", "7"));
      v.add ( new ConfigProperty ("warning", "Days before competence deadline to warn", "2"));
      return v;
   }

}
