/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.raspberry;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.ucam.srcf.assassins.ConfigProperty;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.PlayerViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginConnector;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.xml.Config;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Raspberry extends Plugin
{
   public static String getDescription() { return "Manages the auto-resurrection of players"; };
   /** A reference to allow us to change the game */
   private PluginConnector vanilla;
   /** the amount to increase competence by */
   private int defaultTimeExtension;
   
   public void setPluginConnector(PluginConnector p)
   {
      super.setPluginConnector(p);
      String temp = Config.getConfig(this.getPluginName(), "defaultTimeExtension");
      defaultTimeExtension = Integer.parseInt(temp);
   }

   /**
    * No initialisation neccessary
    */
   protected  void gameCreated() { return; }
   
   /**
    * Add a resurrection event later automatically
    */
   protected  void eventAdded(EventViewer e, Vector qs) throws AssassinSQLException
   { 
      long t = e.getTime();
      Event n = new Event();
      n.setEventTime(t + defaultTimeExtension);
      n.setVisible(false);
      List l = e.getStateChanges();
      Iterator i = l.iterator();
      
      while (i.hasNext()) {
         StateChange s = (StateChange) i.next();
         if (s instanceof Death) {
            n.addStateChange(new Resurrection(s.getAssassinViewer()));
         }
      }
   }
   
   /**
    * Do nothing here.
    */
   protected  void eventModified(EventViewer e, Vector qs) { return ; } 
   
   /**
    * Do nothing here
    */
   protected  void updateState() { return; }
   
   /**
    * Has no properties
    */
   protected  boolean playerHasProperty(PlayerViewer p, String prop) { return false; }

   /**
    * Never need to email people
    */
   protected  boolean playerNeedsEmail(PlayerViewer p) { return false; }

   /**
    * Never need to send any email
    */
   protected  String generateEmail(PlayerViewer p) { return null; }

   /**
    * Never asks questions.
    */
   protected  Vector getQuestions(EventViewer e) { return new Vector(); }
}


