/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.tuttifrutti;

import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.HashMap;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.OptionQuestion;
import org.ucam.srcf.assassins.PlayerViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginCommand;
import org.ucam.srcf.assassins.PluginConnector;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.Question;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.TextQuestion;
import org.ucam.srcf.assassins.YesNoQuestion;
/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class TargettingPlugin extends Plugin
{
   class SCComp implements Comparator
   {
      public int compare(Object sc1, Object sc2) {
         long t1 = 0;
         long t2 = 0;
         try {
            t1 = ((StateChange) sc1).getEventViewer().getTime();
            t2 = ((StateChange) sc2).getEventViewer().getTime();
         } catch (AssassinSQLException asqle) {
            // Do nothing
         }
         if (t1 > t2) return -1;
         if (t2 > t1) return 1;
         if ((sc1 instanceof Death) && (sc2 instanceof Resurrection)) return -1;
         if ((sc2 instanceof Death) && (sc1 instanceof Resurrection)) return 1;
         return 0;	
      }

      public boolean equals(Object sc1, Object sc2) throws AssassinSQLException {
         return (compare(sc1, sc2) == 0);
      }
   }

   private static String formatPlayerInfo(PlayerViewer p) throws AssassinSQLException {
      return "Name: " + p.getName() + "\nCollege: " + p.getCollege() + "\nAddress: " + p.getAddress() + "\nWater Weapons Status: " + p.getWaterStatus() + "\nNotes: " + p.getNotes() + "\n";
   }
   private TargettingAlgorithm buildState(Map state) throws AssassinSQLException {   
      if (Debug.debug) Debug.print(TargettingPlugin.class, "in buildState");
      Set aset = state.keySet();
      List sclist = StateChange.getProcessedAtTime(false, vanilla.getNowTime());
      if (Debug.debug) Debug.print(TargettingPlugin.class, "got "+sclist.size()+" state changes");
      TargettingAlgorithm tAlg = new TFAlgorithm(this, state);
      if (Debug.debug) Debug.print(TargettingPlugin.class, "sorting state changes");
      if (Debug.debug) Debug.print(TargettingPlugin.class, "sorted");
      Iterator i = aset.iterator();
      while (i.hasNext()) {
         if (Debug.debug) Debug.print(TargettingPlugin.class, "iterating assassins");
         AssassinViewer a = (AssassinViewer) i.next();
         if (Debug.debug) Debug.print(TargettingPlugin.class, "got " + a);
         if (((AssassinStateViewer) state.get(a)).getBoolProperty(getPluginName(), "in use")) {
            if (Debug.debug) Debug.print(TargettingPlugin.class, "adding "+a);
            tAlg.addAssassin(a);
         }
      }
      i = sclist.iterator(); 
      while (i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         if (Debug.debug) Debug.print(TargettingPlugin.class, "Found SC: "+sc);
         if (sc instanceof PluginStateChange) {
            if (sc.getPluginName().equals(getPluginName())) {
               String field = sc.getFieldName();
               if (Debug.debug) Debug.print(TargettingPlugin.class, "SC: "+field);
               if (field.equals("entarget")) {
                  if (Debug.debug) Debug.print(TargettingPlugin.class, "giving target"+sc.getEntityViewer());
                  tAlg.giveTarget(sc.getAssassinViewer(), (AssassinViewer) sc.getEntityViewer());
               }
               if (field.equals("detarget")) {
                  tAlg.takeTarget(sc.getAssassinViewer(), (AssassinViewer) sc.getValue());
               }
            }
         }
         if (sc instanceof Death) {
            if (Debug.debug) Debug.print(TargettingPlugin.class, "removing "+sc.getAssassinViewer());
            tAlg.removeAssassin(sc.getAssassinViewer());
         }
         if (sc instanceof Resurrection) {
            AssassinViewer a = ((Resurrection) sc).getAssassinViewer();
            if (aset.contains(a)) {
               tAlg.addAssassin(a);
            }
         }
      }
      if (Debug.debug) Debug.print(TargettingPlugin.class, "state built");
      return tAlg;
   }
   public TargettingPlugin() {
   }
   public void setPluginConnector(PluginConnector p) {
      this.vanilla = p;
   }
   /**
    * Returns a canonical textual string representing the Plugin
    * @return The canonical textual string representing the Plugin.
    */
   public String getPluginName() {
      return "tuttifrutti";
   }

   /**
    * Called whenever the event to which an existing state change is
    * attached has its time changed.
    * @param sc The state change affected.
    * @param oldtime The previous timing of the state change's associated
    *    event.
    * @param newtime The new timing of the event. Note that this can also
    *    be retrieved with {@link StateChange#getEvent} and
    *    {@link Event#getTime}, but is provided here for convenience.
    */
   protected void stateChangeTimeChanged(StateChange sc, int oldtime, int newtime) {}

   /**
    * Called whenever the plugin is required to update or sanity-check the
    * current gamestate.
    */
   protected Vector updateState(Map state, Vector qs, EventViewer e) throws AssassinSQLException {
      if (Debug.debug) Debug.print(TargettingPlugin.class, "in updateState");
      Set aset = state.keySet();
      TargettingAlgorithm tAlg = new TFAlgorithm(this, state);
      Vector livelist = new Vector();
      Vector dlist = new Vector();
      Iterator i = aset.iterator();
      while (i.hasNext()) {
         AssassinViewer a = (AssassinViewer) i.next();
         if (((AssassinStateViewer) state.get(a)).getBoolProperty(getPluginName(), "in use")) {
            tAlg.addAssassin(a);
            if (((AssassinStateViewer) state.get(a)).isAlive())
               livelist.add(a);
            else
               dlist.add(a);
         }
      }
      livelist.addAll(dlist);

      if (Debug.debug) Debug.print(TargettingPlugin.class, "building state");
      tAlg = buildState(state);
      if (qs.size() > 0) {

         i = qs.iterator();
         while (i.hasNext()) {
            OptionQuestion q = (OptionQuestion) i.next();
            vanilla.addStateChange(e, new PluginStateChange((AssassinViewer) q.getHidden(), this, "entarget", (AssassinViewer) livelist.get(q.getAnswerIndex())));
            vanilla.addStateChange(e, new PluginStateChange((AssassinViewer) q.getHidden(), this, "hasnewtargets", ""));
         }
         return null;
      }

      List deadlist = new LinkedList();
      List resdlist = new LinkedList();
      List sclist = StateChange.getProcessedAtTime(false, vanilla.getNowTime());
      if (Debug.debug) Debug.print(TargettingPlugin.class, "got unprocessed list");
      /*Collections.sort(sclist, new SCComp());*/
      i = sclist.iterator();
      while (i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         if (Debug.debug) Debug.print(TargettingPlugin.class, "iterating state changes");
         if (sc instanceof Death) {
            // if (Debug.debug) Debug.print(TargettingPlugin.class, "found a death: " + sc.getAssassin());
            AssassinViewer v = ((Death) sc).getVictim();
            if (Debug.debug) Debug.print(TargettingPlugin.class, "got a victim: " + v);
            if (resdlist.contains(v)) {
               resdlist.remove(v);
            }
            else {
               if (Debug.debug) Debug.print(TargettingPlugin.class, "adding a death");
               deadlist.add(v);
            }
         }
         if (sc instanceof Resurrection) {
            AssassinViewer r = ((Resurrection) sc).getAssassinViewer();
            if (deadlist.contains(r)) {
               deadlist.remove(r);
            }
            else resdlist.add(r);
         }
      }
      if (resdlist.size() > 0) {
         i = deadlist.iterator();
         while (i.hasNext()) {
            tAlg.removeAssassin((AssassinViewer) i.next());
         }
         i = resdlist.iterator();
         while (i.hasNext()) {
            tAlg.addAssassin((AssassinViewer) i.next());
         }
         sclist = tAlg.reallocateTargets();
      }
      else {
         sclist = new LinkedList();
         i = deadlist.iterator();
         while (i.hasNext()) {
            AssassinViewer a = (AssassinViewer) i.next();
            if (Debug.debug) Debug.print(TargettingPlugin.class, "removing an assassin: " + a);
            tAlg.removeAssassin(a);
            if (Debug.debug) Debug.print(TargettingPlugin.class, "removed");
            sclist.addAll(tAlg.reallocateTargets());
            if (Debug.debug) Debug.print(TargettingPlugin.class, "got reallocations");

         }
         if (Debug.debug) Debug.print(TargettingPlugin.class, "Final reallocation");
         sclist.addAll(tAlg.reallocateTargets());
         if (Debug.debug) Debug.print(TargettingPlugin.class, "Succeeded!");
      }

      Vector qlist = new Vector();
      i = sclist.iterator();
      while (i.hasNext()) {
         PluginStateChange sc = (PluginStateChange) i.next();
         if (sc.getFieldName().equals("entarget")) {
            if (Debug.debug) Debug.print(TargettingPlugin.class, "Questioning "+sc.getAssassinViewer()+" -> "+sc.getEntityViewer());
            OptionQuestion q = new OptionQuestion("Assign as target for " + sc.getAssassinViewer().getPlayerViewer().getName(), livelist, livelist.indexOf(sc.getEntityViewer()));
            q.setHidden(sc.getAssassinViewer());
            qlist.add(q);
            if (Debug.debug) Debug.print(TargettingPlugin.class, "adding question");
         }
         else {
            vanilla.addStateChange(e, sc);
         }
      }
      if (qlist.size() == 0) return null;
      return qlist;
   }

   /**
    * Called during the updating of the gamestate to ask whether a given player
    * requires email to be sent to them during the current round of emails.
    * @param p A viewer for the player being asked about.
    * @return True if the plugin requires email to be sent to the player in
    *    question, false if not.
    */
   protected boolean playerNeedsEmail(AssassinStateViewer asv) throws AssassinSQLException {
      AssassinViewer av = asv.getAssassin();
      List sclist = StateChange.getByFieldAtTime("hasnewtargets", vanilla.getNowTime());
      boolean ans = false;
      Iterator i = sclist.iterator();
      while (i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         if (sc.getAssassinViewer().equals(av)) {
            ans = true;
            vanilla.removeStateChange(sc.getEventViewer(), sc);
         }
      }
      return ans;
   }

   /**
    * Called when the system has decided that email needs to be sent to a given
    * player. The plugin may return a string to be included in the email, or
    * null if it does not wish to include anything.
    * @param p A viewer for the player being emailed.
    * @return A string to be included in the email sent to the player, or null.
    *    The string will appear in its own paragraph when inserted into the
    *    email and so should not begin or end with a newline.
    */
   protected String generateEmail(AssassinStateViewer asv, Map state) throws AssassinSQLException {
      if (false == asv.getBoolProperty(getPluginName(), "in use")) return null;
      if (Debug.debug) Debug.print(TargettingPlugin.class, "in generateEmail");
      if (Debug.debug && (null == asv)) Debug.print(TargettingPlugin.class, "FISH!");
      TargettingAlgorithm tAlg = buildState(state);
      if (Debug.debug) Debug.print(TargettingPlugin.class, "state built");
      AssassinViewer av = asv.getAssassin();
      if (Debug.debug) Debug.print(TargettingPlugin.class, "got assassin");
      List targlist = tAlg.getTargets(av);
      if (Debug.debug) Debug.print(TargettingPlugin.class, "got targets");
      String email = "Your targets are:\n\n";
      Iterator i = targlist.iterator();
      while (i.hasNext()) {
         if (Debug.debug) Debug.print(TargettingPlugin.class, "iterating targets");
         AssassinViewer blob = (AssassinViewer) i.next();
         if (((AssassinStateViewer) state.get(blob)).isAlive()) {
            email += formatPlayerInfo((blob.getPlayerViewer())) + "\n";
         }
      }
      return email;
   }

   /**
    * Called when an event is being added to get any queries the plugin has for
    * the user. The plugin should check all the event details and if it needs
    * more information it should ask relevent questions. The eventAdded 
    * function is passed the same Vector with the answers filled out.
    * @param e The event which is being added
    * @return A Vector of Questions
    * @see Question
    * @see Plugin.eventAdded
    */
   protected Vector getQuestions(EventViewer e) {return new Vector();}
   protected Vector getQuestions(AssassinViewer a, Map state) throws AssassinSQLException {
      if (Debug.debug) Debug.print(TargettingPlugin.class, "Asking questions");
      Vector v = new Vector();
      Question seedq = new TextQuestion("Seed value", "0");
      Question targq = new YesNoQuestion("Use targetting?", true);
      v.add(seedq);
      v.add(targq);
      return v;
   }
   protected void returnQuestions(AssassinViewer a, EventViewer e, Map state, Vector qs) throws AssassinSQLException {
      int seed = Integer.parseInt(((TextQuestion) qs.get(0)).getAnswer());
      boolean inuse = ((YesNoQuestion) qs.get(1)).getAnswer();
      StateChange seedsc = new PluginStateChange(a, this, "seed", String.valueOf(seed));
      StateChange targsc = new PluginStateChange(a, this, "in use", String.valueOf(inuse));
      vanilla.addStateChange(e, seedsc);
      vanilla.addStateChange(e, targsc);
   }
   public void dumpdot(Map params) throws AssassinSQLException, StateBuilderException{
      Map state = vanilla.getState();
      TargettingAlgorithm tAlg = buildState(state);
      System.out.print("digraph Targetting {\n");
      for (Iterator i = state.keySet().iterator(); i.hasNext();) {
         AssassinViewer a = (AssassinViewer) i.next();
         AssassinStateViewer as = (AssassinStateViewer) state.get(a);
         if (as.isAlive() && as.getBoolProperty(getPluginName(), "in use")) {
            List tList = tAlg.getTargets(a);
            for (Iterator j = tList.iterator(); j.hasNext();) {
               AssassinViewer b = (AssassinViewer) j.next();
               AssassinStateViewer bs = (AssassinStateViewer) state.get(b);
               if (bs.isAlive() && bs.getBoolProperty(getPluginName(), "in use")) {
                  PlayerViewer p = a.getPlayerViewer();
                  PlayerViewer q = b.getPlayerViewer();
                  System.out.print("\"" + p.getName() + "\" -> \"" + q.getName() + "\";\n");
               }
            }
         }
      }
      System.out.print("}\n");
   }

   public Map getCommands()
   {
      Map m = new HashMap();
      final TargettingPlugin tp = this;
      m.put("dump-dot",
            new PluginCommand() {
               public void pluginCommand(Map params) throws AssassinSQLException, StateBuilderException
               {
                  tp.dumpdot(params);
               }
            }
           );
      return m;
   }

}
